local tasks = {sleeping = 0}

local max_task_switches_per_frame = 50

local isdown = love.keyboard.isScancodeDown

-- This is intended for setfenv, but it's not used that way in this example.
local api = {
  -- Functions that trigger a task switch
  wait = function (time)
    coroutine.yield("wait", time)
  end;
  waitkeydown = function (key, timeout)
    coroutine.yield("keydown", key, timeout)
  end;
  waitkeyup = function (key, timeout)
    coroutine.yield("keyup", key, timeout)
  end;
}

-- Wake-up events
local function wake_up_by_time(time)
  return love.timer.getTime() >= time
end

local function wake_up_by_keydown(key, timeout)
  if isdown(key) then return true end
  if timeout then return love.timer.getTime() >= timeout end
end

local function wake_up_by_keyup(key, timeout)
  if not isdown(key) then return true end
  if timeout then return love.timer.getTime() >= timeout end
end


-- Implementation of the API functions proper. The entries are the strings
-- passed as the first parameter of coroutine.yield().
-- In this case, they all set the tasks to sleep and set up a wake-up event.
local dispatchers = {
  wait = function (task, time)
    task.wake_up_event = {wake_up_by_time, love.timer.getTime() + time}
  end;
  keydown = function(task, key, timeout)
    task.wake_up_event = {wake_up_by_keydown, key, timeout and timeout + love.timer.getTime()}
  end;
  keyup = function(task, key, timeout)
    task.wake_up_event = {wake_up_by_keyup, key, timeout and timeout + love.timer.getTime()}
  end;
}


local function add_task(fn, userdata, priority)
  table.insert(tasks, {
    co = coroutine.create(fn);
    userdata = userdata;
    priority = priority or 1;
    pricnt = 0;
    wake_up_event = false;
  })
end


local function dispatch(task, status, fn, ...)
  -- Functions that terminate return nil; check for that here
  if fn ~= nil then
    local old_event = task.wake_up_event

    dispatchers[fn](task, ...)

    if old_event and not task.wake_up_event then
      tasks.sleeping = tasks.sleeping - 1
    elseif not old_event and task.wake_up_event then
      tasks.sleeping = tasks.sleeping + 1
    end
  end
end


local function scheduler()
  for i = 1, #tasks do
    tasks[i].checked = false
  end
  -- Ensure we check every task at least once
  local to_check = #tasks
  local i = 0
  local inc = 1
  local task_switches = 0
  repeat

    i = i + inc
    inc = 1
    if i > #tasks then
      if #tasks == 0 then
        break
      end
      i = 1
    end
    local task = tasks[i]

    -- Handle priority
    task.pricnt = task.pricnt + 1
    if task.pricnt >= task.priority then
      task.pricnt = 0
    end

    -- Check if this task needs to be run.
    -- If an even wakes it up, it will be run regardless of priority,
    -- but its priority counter will be reset.
    local needs_run = task.wake_up_event == false and task.pricnt == 0
    if not needs_run then
      needs_run = task.wake_up_event
      if needs_run then
        needs_run = needs_run[1](select(2, unpack(task.wake_up_event)))
        if needs_run then
          task.wake_up_event = false
          tasks.sleeping = tasks.sleeping - 1
        end
        if not task.checked then
          to_check = to_check - 1
          task.checked = true
        end
      end
    end
    if needs_run then
      task.pricnt = 0
      task_switches = task_switches + 1

      dispatch(task, assert(coroutine.resume(task.co, task.userdata)))

      task.userdata = false
      if coroutine.status(task.co) == "dead" then
        -- Destroy task
        tasks[i] = tasks[#tasks]
        tasks[#tasks] = nil
        if task.wake_up_event ~= false then
          -- this task was sleeping, but we've deleted it - decrease count
          tasks.sleeping = tasks.sleeping - 1
        end
        if not task.checked then
          -- this task was still counted in the to_check count, update
          to_check = to_check - 1
        end
        task = nil -- kill object ASAP
        inc = 0
      end
    end
  until to_check == 0 and (tasks.sleeping == #tasks or task_switches >= max_task_switches_per_frame)

end


local function script_randomguy(obj)
  obj.x = love.math.random(love.graphics.getWidth())
  obj.y = love.math.random(love.graphics.getHeight())
  repeat
    api.wait(0.1)
    obj.x = obj.x + love.math.random(-1, 1)
    obj.y = obj.y + love.math.random(-1, 1)
    if obj.x < 0 then obj.x = 0 end
    if obj.y < 0 then obj.y = 0 end
    if obj.x > love.graphics.getWidth()-20 then obj.x = love.graphics.getWidth()-20 end
    if obj.y > love.graphics.getHeight()-20 then obj.y = love.graphics.getHeight()-20 end
  until false
end


local function script_kbguy(obj)
  repeat
    if isdown"w" then
      obj.y = obj.y - 5
    end
    if isdown"s" then
      obj.y = obj.y + 5
    end
    if isdown"a" then
      obj.x = obj.x - 5
    end
    if isdown"d" then
      obj.x = obj.x + 5
    end
    api.wait(0.02)
  until false
end


local skip = false

local function script_textinfo(obj)
  local text = "Lorem ipsum dolor sit amet, consectetur adipiscing elit."
    .. " Nullam eu varius metus. Fusce ex sem, consectetur sit amet ullam"
    .. " sit amet, maximus vitae nunc. Curabitur pretium malesuada dui ac"
    .. " ornare. Praesent in dolor tempor, condimentum sem vel, commodo"
    .. " justo. Donec pharetra, magna nec viverra gravida, enim velit"
    .. " imperdiet ante, ut viverra nibh velit quis neque. In quam nunc,"
    .. " lobortis nec ex quis, mattis porttitor enim. Nulla ac arcu in"
    .. " lacus cursus rutrum. Vivamus ornare magna non dolor pellentesque,"
    .. " et mattis ipsum molestie. Quisque imperdiet libero arcu, a"
    .. " fermentum nunc consequat quis. Sed nisi augue, accumsan et turpis"
    .. " vitae, imperdiet vehicula tortor. Duis at dui arcu. Ut feugiat"
    .. " aliquam sapien id blandit. "
  obj.text = ""
  obj.w = 400
  local font = love.graphics.getFont()
  local width, lines = font:getWrap(text, obj.w)
  -- This avoids the last word jumping to the next line as it's typed.
  text = table.concat(lines, "\n")
  local textlen = #text
  obj.y = love.graphics.getHeight() - #lines * font:getHeight()
  local progress = 0
  repeat
    obj.text = text:sub(1, progress)
    repeat
      progress = progress + 1
    until text:byte(progress) ~= 10 -- skip over newlines
    if skip then
      progress = textlen
      obj.text = text
    else
      api.wait(isdown("space") and 0.015 or 0.04)
    end
  until progress >= textlen
  -- prevent accident of pressing enter just as the text finishes
  if not skip then api.wait(0.5) end
  api.waitkeyup("return")
  api.waitkeydown("return")
  obj.text = ""
end


-- Demo
local objs = {}

function love.load()
  for i = 1, 5 do
    local obj = { type = "sprite" }
    table.insert(objs, obj)
    add_task(script_randomguy, obj)
  end
  do
    local obj = { type = "text"; x = 0; y = love.graphics.getHeight() - 100; }
    table.insert(objs, obj)
    add_task(script_textinfo, obj)
  end
  do
    local obj = { type = "sprite"; x = love.graphics.getWidth()*.5; y = love.graphics.getHeight()*.5; red = true; }
    table.insert(objs, obj)
    add_task(script_kbguy, obj)
  end
end

function love.keypressed(key, scan)
  if scan == "return" then skip = true end
  return scan == "escape" and love.event.quit()
end

function love.update(dt)
  scheduler()
end

function love.draw()
  for i = 1, #objs do
    local obj = objs[i]
    love.graphics.setColor(1, obj.red and 0 or 1, obj.red and 0 or 1)
    if obj.type == "sprite" then
      love.graphics.rectangle("fill", obj.x, obj.y, 20, 20)
    elseif obj.type == "text" then
      love.graphics.printf(obj.text, obj.x, obj.y, obj.w)
    end
  end
end
